/* topology.cc
 * Copyright (C) 2001 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include "topology.h"
#include "scheduler.h"
#include "discover.h"
#include "preferences.h"
#include "process.h"
#include "task.h"
#include "main.h"
#include "session.h"
#include "data.h"
#include "misc.h"
#include "gui.h"
#include "parser.h"
#include "monitor.h"
#include "ref.h"

Topology::Topology(Main *m)
{
	main = m;
	parser = new Parser();
	default_network_type = network_auto;
}

Topology::~Topology()
{
	if (parser!=NULL) delete parser;
}

void Topology::create_placeholder(char *nmstr, IPAddress tip)
{
	Device *d;

	if ((!tip.zero()) && (!tip.loopback()))
	{
		// see if device already exists
		d = main->session->data->find_ip_on_device(tip);
		if (d==NULL)
		{
			d = new Device();
			d->pingname.set(tip);
			strcpy(d->name, nmstr);
			d->maintype = visual_device;
			d->subtype = device_placeholder;
			main->session->data->devices.tag(d);
			main->gui->update_views();
			main->gui->update_lists();
		}
	}
}

void Topology::deviceAdd(char *fnameparam, int hop)
{
	char buf[256], fname[256];
	GError *error;
	gboolean ch;

	// parse the SNMP data
	sscanf(fnameparam, "%[^ ] %[^\n]", fname, buf);
	sprintf(buf, "Parsing %s", fname);
	main->gui->monitor->show_message(buf);
	sprintf(buf, "%s/topology/%s", main->session->datadir, fname);
	parser->load_device(buf, fname);
	if (parser->d==NULL) return;

	// create new device
	tmpdev = new Device();
	origdev = NULL;

	// defaults for new device
	tmpdev->grab_top = (gconf_client_get_bool(main->client,
		CONFIG_SESSION_TOPOLOGY_DEFAULT, &error)==TRUE)?'1':'0';
	tmpdev->grab_stats = (gconf_client_get_bool(main->client,
		CONFIG_SESSION_GRAPH_DEFAULT, &error)==TRUE)?'1':'0';

	copyDeviceHeader();							// copy over device header from parser
	clearDeviceTags();							// deletion tags
	checkForDuplicateDevice();			// identify device that is being replaced
	mergeType();
	attachInterfaces();							// create interfaces & networks
	addRef();
	addNextHop();

	// at this point we have a fully populated device
	main->session->data->devices.tag(tmpdev);
	if (origdev==NULL) main->session->devicebit = TRUE;
	deleteTaggedDevices();

	layer3(hop);										// lets attach at layer 3
	layer2hack();
//	layer2(tmpdev);												// now attach at layer 2
//	layer2go();
	copyOverMachines();							// create end user machine entries
	setPreferences();

	main->gui->update_views();
	main->gui->update_lists();
}

void Topology::deviceDeleteSelected()
{
	Link *ln;
	Device *d, *s;
	gboolean quit = FALSE;

	printf("BEFORE DO DELETE SELECTED\n");

	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;

		printf("Device %s\n", d->name);
	}
	printf("OK NOW DO DELETIONS\n");


	while(!quit)
	{
		// scan for any devices in need of deleting
		d = NULL;
	 	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
		{
			s = (Device*) ln->item;
			if ((s!=NULL) && (s->selection==TRUE)) d = s;
		}
		if (d != NULL)
		{
//			layer2(d);
			main->session->data->devices.del(d);
			deleteSingleDevice(d);
//			layer2go();
			tmpdev=d;			// test hack for layer2hack();
			layer2hack();
			if (!main->session->devicebit)
			{
				main->session->devicebit=TRUE;
				main->setTitle();
			}
		}else
			quit = TRUE;
	}

	main->gui->update_views();
	main->gui->update_lists();
}

void Topology::deviceDelete(Device *d)
{
	layer2(d);
	main->session->data->devices.del(d);
	deleteSingleDevice(d);
	layer2go();
	if (!main->session->devicebit)
	{
		main->session->devicebit=TRUE;
		main->setTitle();
	}

printf("Device deleted\n");

	main->gui->update_views();
	main->gui->update_lists();
}

// we rely on the user merging the right devices,
// but even if they don't we should not crash
void Topology::deviceMerge()
{
	Device *d, *d1, *d2;
	Link *ln, *ln2;
	Interface *in, *in2, *tin;
	IPAddress *ip;
	Network *n;
	Machine *m;
	Ref *r;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if (d->selection==TRUE)
		{
			if (d->subtype==device_router) d1 = d;
			if (d->subtype==device_switch) d2 = d;
		}
	}

	clearDeviceTags();
	d1->tag = 1;
	d2->tag = 1;

	printf("MERGE the two devices %s and %s\n", d1->name, d2->name);

	tmpdev = new Device();
	tmpdev->maintype = visual_device;
	tmpdev->subtype = device_merged;
	tmpdev->pingname = d1->pingname;
	tmpdev->pingname2 = d2->pingname;
	strcpy(tmpdev->name, d1->name);
	tmpdev->last_seen = d1->last_seen;
	strcpy(tmpdev->location, d1->location);
	strcpy(tmpdev->community, d1->community);
	strcpy(tmpdev->community2, d2->community);
	strcpy(tmpdev->descr, d1->descr);
	strcpy(tmpdev->comment, d1->comment);
	strcpy(tmpdev->contact, d1->contact);
	tmpdev->status = d1->status;
	tmpdev->group = d1->group;
	tmpdev->grab_top = d1->grab_top;
	tmpdev->grab_stats = d1->grab_stats;

	tmpdev->x = d1->x;
	tmpdev->y = d1->y;
	tmpdev->selection = d1->selection;
	tmpdev->shown = d1->shown;

	for (ln=d2->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;

		tin = new Interface();
		tin->index = in->index;
		tin->speed = in->speed;
		tin->mac = in->mac;
		tin->type = in->type;
		tin->status = in->status;
		strcpy(tin->descr, in->descr);
		strcpy(tin->comment, in->comment);
		for (ln2=in->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			ip = (IPAddress*) ln2->item;
			if ((ip != NULL) && (!ip->zero()))
				addIPToIPAddresses(tin, *ip);
		}

		for (ln2=in->refs.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			r = (Ref*) ln2->item;
			tin->refs.tag(r);
			ln2->item = NULL;
		}
		in->refs.tidy();

		for (ln2=in->machines.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			m = (Machine*) ln2->item;
			if (m != NULL)
			{
				m->interface = tin;
				tin->machines.tag(m);
				ln2->item = NULL;
			}
		}
		in->machines.tidy();

		in2 = lookupInterface(d1, in);
		if (in2!=NULL)
		{
			tin->speed = in2->speed;
			tin->type = in2->type;
			tin->status = in2->status;
			strcpy(tin->comment, in2->comment);
			for (ln2=in2->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				ip = (IPAddress*) ln2->item;
				if ((ip != NULL) && (!ip->zero()))
					addIPToIPAddresses(tin, *ip);
			}

			for (ln2=in2->refs.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				r = (Ref*) ln2->item;
				tin->refs.tag(r);
				ln2->item = NULL;
			}
			in->refs.tidy();

			for (ln2=in2->nexthops.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				ip = (IPAddress*) ln2->item;
				if ((ip != NULL) && (!ip->zero()))
				{
					tin->nexthops.tag(ip);
					ln2->item = NULL;
				}
			}
			in2->nexthops.tidy();

			for (ln2=in2->machines.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				m = (Machine*) ln2->item;
				if (m != NULL)
				{
					m->interface = tin;
					tin->machines.tag(m);
					ln2->item = NULL;
				}
			}
			in2->machines.tidy();

			n = in2->network;
			n->interfaces.tag(tin);
			if (default_network_type==network_auto) n->subtype = network_auto;
			tin->network = n;
		}

		if (tin->network==NULL)
		{
			n = new Network();
			n->interfaces.tag(tin);
			if (default_network_type==network_auto) n->subtype = network_auto;
			tin->network = n;
			main->session->data->networks.tag(n);
		}

		tin->device = tmpdev;
		tmpdev->interfaces.tag(tin);
	}


	// using routers networks so already attached at layer3
//	layer2(tmpdev);
//	layer2go();
	layer2hack();
	setPreferences();

	main->session->data->devices.tag(tmpdev);
	deleteTaggedDevices();

	main->session->devicebit = TRUE;

	main->gui->update_views();
	main->gui->update_lists();
}

void Topology::deviceSplit(Device *d)
{
	char buf[256];

	printf("SPLIT\n");

	printf("We need to split the device into %s and ", d->pingname.getstring());
	printf("%s\n", d->pingname2.getstring());
	// We also need to preserve all user defined data & end user machines

	// tag the merged device for deletion
	clearDeviceTags();
	d->tag = 1;

	// parse the router side
	sprintf(buf, "%s/topology/%s",
		main->session->datadir, d->pingname.getstring());
	parser->load_device(buf, d->pingname.getstring());
	if (parser->d==NULL) return;

	tmpdev = new Device();
	origdev = parser->d;
	copyDeviceHeader();				// copy over device header from parser
	tmpdev->x = d->x;
	tmpdev->y = d->y;
	tmpdev->selection = d->selection;
	tmpdev->shown = d->shown;
	merge_type = 1;						// we are creating a new router (not replacing old)
	attachInterfaces();				// create interfaces & networks
	layer3(0);								// lets attach at layer 3
//	layer2(tmpdev);									// now attach at layer 2
//	layer2go();
	layer2hack();
	//	copyOverMachines();				// create end user machine entries
	setPreferences();
	main->session->data->devices.tag(tmpdev);

	// parse the switch side
	sprintf(buf, "%s/topology/%s",
		main->session->datadir, d->pingname2.getstring());
	parser->load_device(buf, d->pingname2.getstring());
	if (parser->d==NULL) return;

	tmpdev = new Device();
	origdev = parser->d;
	copyDeviceHeader();				// copy over device header from parser
	merge_type = 2;						// we are creating a new router (not replacing old)
	// new position for switch half
	tmpdev->selection = d->selection;
	tmpdev->shown = d->shown;
	attachInterfaces();				// create interfaces & networks
	layer3(0);								// lets attach at layer 3
//	layer2(tmpdev);									// now attach at layer 2
//	layer2go();
	layer2hack();
	//	copyOverMachines();				// create end user machine entries
	setPreferences();
	main->session->data->devices.tag(tmpdev);

	deleteTaggedDevices();

	main->session->devicebit = TRUE;

	main->gui->update_views();
	main->gui->update_lists();
}

/////////////////////////////////////////////////////////////////////////////
// Private Classes
/////////////////////////////////////////////////////////////////////////////

void Topology::setPreferences()
{
	Link *ln, *ln2;
	Interface *in;
	Machine *m;
	gchar *str;
	gboolean ch1, ch2, ch3;
	GError *error;
	NetworkType type1;
	NetworkPositionType type2;
	NetworkGroupType type3;

	// only set defaults if creating new entries
	// otherwise keep existing entries
	if ((merge_type>=1) && (merge_type<=2))
	{
		str = gconf_client_get_string(main->client,
			CONFIG_DIAGRAM_GROUP_LEAF, &error);
		tmpdev->group = stringToDeviceGroupType(str);
		tmpdev->grab_top = (gconf_client_get_bool(main->client,
			CONFIG_SESSION_TOPOLOGY_DEFAULT, &error)==TRUE)?'1':'0';
		tmpdev->grab_stats = (gconf_client_get_bool(main->client,
			CONFIG_SESSION_GRAPH_DEFAULT, &error)==TRUE)?'1':'0';
		ch1 = gconf_client_get_bool(main->client,
			CONFIG_DIAGRAM_SHOW_LEAF, &error);
		ch2 = gconf_client_get_bool(main->client,
			CONFIG_DIAGRAM_SHOW_EMPTY_LEAF, &error);
		ch3 = gconf_client_get_bool(main->client,
			CONFIG_DIAGRAM_SHOW_MACHINES, &error);
	printf("b1=%s, b2=%s, b3=%s\n", (ch1==TRUE)?"T":"F", (ch2==TRUE)?"T":"F", (ch3==TRUE)?"T":"F");

		str = gconf_client_get_string(main->client,
			CONFIG_DIAGRAM_NETWORK_TYPE, &error);
		type1 = stringToNetworkType(str);
		str = gconf_client_get_string(main->client,
			CONFIG_DIAGRAM_NETWORK_POSITION, &error);
		type2 = stringToNetworkPositionType(str);
		str = gconf_client_get_string(main->client,
			CONFIG_DIAGRAM_GROUP_MACHINES, &error);
		type3 = stringToNetworkGroupType(str);

		for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			if (in->network->interfaces.size()==1)
			{
				in->network->shown = FALSE;
				if (ch1==TRUE)
				{
					in->network->shown = TRUE;
					if ((ch2==FALSE) && (in->machines.size()==0))
						in->network->shown = FALSE;
				}
			}
			in->network->subtype = type1;
			in->network->position = type2;
			in->network->group = type3;
			for (ln2=in->machines.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				m = (Machine*) ln2->item;
				m->shown = ch3;
			}
		}
	}
}

// copy over stuff from parser->d
void Topology::copyDeviceHeader()
{
	tmpdev->maintype = parser->d->maintype;
	tmpdev->subtype = parser->d->subtype;
	tmpdev->pingname = parser->d->pingname;
	tmpdev->pingname2 = parser->d->pingname2;
	strcpy(tmpdev->name, parser->d->name);
	tmpdev->last_seen = parser->d->last_seen;
	strcpy(tmpdev->location, parser->d->location);
	strcpy(tmpdev->community, parser->d->community);
	strcpy(tmpdev->community2, parser->d->community2);
	strcpy(tmpdev->descr, parser->d->descr);
	strcpy(tmpdev->comment, parser->d->comment);
	strcpy(tmpdev->contact, parser->d->contact);
	// default status of unknown
	// default group which we should get from preferences
	// default grab topology
	// default grab statistics
}

void Topology::clearDeviceTags()
{
	Link *ln;
	Device *d;

	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		d->tag = 0;
	}
}

void Topology::checkForDuplicateDevice()
{
	Link *ln, *ln2;
	Interface *in;
	IPAddress *tip;

	checkForDuplicateDeviceWithIP(&parser->d->pingname);
	for (ln=parser->d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		for (ln2=in->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			tip = (IPAddress*) ln2->item;
			checkForDuplicateDeviceWithIP(tip);
		}
	}
}

void Topology::checkForDuplicateDeviceWithIP(IPAddress *ip)
{
	Link *ln;
	Device *d;

	if (ip->zero()) return;

	// does the IP match a device
	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if ((d->pingname.compare(*ip)==0) ||
			 	(d->pingname2.compare(*ip)==0))
		{
			// printf("FOUND MATCH WITH IP's %s\n", ip->getstring());
			// we don't care if placeholder or real device, just copy
			// coordinates and tag for deletion
			if (d->subtype != device_placeholder) origdev = d;
			tmpdev->x = d->x;
			tmpdev->y = d->y;
			tmpdev->selection = d->selection;
			tmpdev->shown = d->shown;
			d->tag = 1;
		}
	}
}

// There are 6 different combinations for merging devices so make
// things easier we use a merge_type variable
void Topology::mergeType()
{
	if ((tmpdev->subtype==device_router) && (origdev==NULL))
		merge_type = 1;
	else
	if ((tmpdev->subtype==device_switch) && (origdev==NULL))
		merge_type = 2;
	else
	if ((tmpdev->subtype==device_router) && (origdev!=NULL) &&
			(origdev->subtype==device_router))
		merge_type = 3;
	else
	if ((tmpdev->subtype==device_switch) && (origdev!=NULL) &&
			(origdev->subtype==device_switch))
		merge_type = 4;
	else
	if ((tmpdev->subtype==device_router) && (origdev!=NULL) &&
			(origdev->subtype==device_merged))
		merge_type = 7;
	else
	if ((tmpdev->subtype==device_switch) && (origdev!=NULL) &&
			(origdev->subtype==device_merged))
		merge_type = 8;
	else
		merge_type = 0;

	if ((merge_type==7) || (merge_type==8)) tmpdev->subtype = device_merged;
}

void Topology::attachInterfaces()
{
	Link *ln, *ln2;
	Interface *in, *tin, *in2;
	Network *n;
	IPAddress *ip;
	Device *d;
	Machine *m;

	// keep user data for devices
	if (origdev!=NULL)
	{
		if (merge_type == 8)
		{
			strcpy(tmpdev->name, origdev->name);
			tmpdev->maintype = visual_device;
			tmpdev->subtype = device_merged;
			tmpdev->pingname2 = tmpdev->pingname;
			tmpdev->pingname = origdev->pingname;
			strcpy(tmpdev->community2, tmpdev->community);
			strcpy(tmpdev->community, origdev->community);
		}else
		if (merge_type == 7)
		{
			tmpdev->pingname2 = origdev->pingname2;
			strcpy(tmpdev->community2, origdev->community2);
		}
		if (strcmp(origdev->location, "")!=0)
			strcpy(tmpdev->location, origdev->location);
		if (strcmp(origdev->descr, "")!=0)
			strcpy(tmpdev->descr, origdev->descr);
		if (strcmp(origdev->comment, "")!=0)
			strcpy(tmpdev->comment, origdev->comment);
		if (strcmp(origdev->contact, "")!=0)
			strcpy(tmpdev->contact, origdev->contact);
		tmpdev->status = origdev->status;
		tmpdev->group = origdev->group;
		tmpdev->grab_top = origdev->grab_top;
		tmpdev->grab_stats = origdev->grab_stats;
	}

	// select source for interface list
	d = (merge_type == 7)?origdev:parser->d;

	for (ln=d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;

		tin = new Interface();
		tin->index = in->index;
		tin->speed = in->speed;
		tin->mac = in->mac;
		tin->type = in->type;
		tin->status = in->status;
		strcpy(tin->descr, in->descr);
		strcpy(tin->comment, in->comment);

		if ((merge_type>=1) && (merge_type<=4))
		{
			if ((tmpdev->subtype==device_switch) ||
					(((tmpdev->subtype==device_router)||(tmpdev->subtype==device_merged)) &&
						(tin->status=='1')))
			{
				for (ln2=in->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					ip = (IPAddress*) ln2->item;
					if ((ip != NULL) && (!ip->zero()) && (!ip->loopback()))
						addIPToIPAddresses(tin, *ip);
				}
			}
		}

		tin->device = tmpdev;
		tmpdev->interfaces.tag(tin);

		// if we are merging then we need to copy over stuff like speed, type,
		// status and IP Addresses

		if ((merge_type == 7) || (merge_type == 8))
		{
			if (merge_type == 7)
				in2 = lookupInterface(parser->d, in);
			else
				in2 = lookupInterface(origdev, in);

			if (in2 != NULL)
			{
				tin->speed = in2->speed;
				tin->type = in2->type;
				tin->status = in2->status;
				strcpy(tin->comment, in2->comment);
				if (tin->status == '1')
				{
					for (ln2=in2->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
					{
						ip = (IPAddress*) ln2->item;
						if ((ip != NULL) && (!ip->zero()) && (!ip->loopback()))
							addIPToIPAddresses(tin, *ip);
					}
				}
			}
		}

		if (tin->network==NULL)
		{
			n = new Network();
			n->interfaces.tag(tin);
			if (default_network_type==network_auto) n->subtype = network_auto;
			tin->network = n;
			main->session->data->networks.tag(n);
		}
	}

//FIXME TOPOLOGY end user machines

	// move over machines from orig, already in global list,
	// moving so no duplicates created
	if ((origdev!=NULL) && (origdev!=parser->d))
	{
		for (ln=origdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			in2 = lookupInterface(tmpdev, in);
			if (in2!=NULL)
			{
				for (ln2=in->machines.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					m = (Machine*) ln2->item;
					in2->machines.tag(m);
					m->interface = in2;
					ln2->item = NULL;
				}
				in->machines.tidy();
			}
		}
	}
}

void Topology::addIPToIPAddresses(Interface *in, IPAddress ip)
{
	IPAddress *ip2;
	Link *ln;

	// scan current list to see if address already exists
	for (ln=in->ip_addresses.get_head(); ln!=NULL; ln=ln->next)
	{
		ip2 = (IPAddress*) ln->item;
		if (ip2->compare(ip)==0)
			return;
	}

	ip2 = new IPAddress();
	ip2->set(ip);
	in->ip_addresses.tag(ip2);
}

void Topology::addIPToNextHop(Interface *in, IPAddress ip)
{
	IPAddress *ip2;
	Link *ln;

	// scan current list to see if address already exists
	for (ln=in->nexthops.get_head(); ln!=NULL; ln=ln->next)
	{
		ip2 = (IPAddress*) ln->item;
		if (ip2->compare(ip)==0)
			return;
	}

	ip2 = new IPAddress();
	ip2->set(ip);
	in->nexthops.tag(ip2);
}

void Topology::addRef()
{
	Link *ln, *ln2;
	Interface *in, *in2;
	Machine *m;
	Ref *r;

	for (ln=parser->d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		in2 = lookupInterface(tmpdev, in);
		if (in2!=NULL)
		{
			for (ln2=in->machines.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				m = (Machine*) ln2->item;
				r = new Ref();
				r->type = parser->d->subtype;
				r->mac.set(m->mac);
				in2->refs.tag(r);
			}
		}
	}

	// if origdev != NULL copy over other half if merged
	if ((merge_type < 7) || (merge_type > 8)) return;

	for (ln=origdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		in2 = lookupInterface(tmpdev, in);
		if (in2!=NULL)
		{
			for (ln2=in->refs.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				r = (Ref*) ln2->item;
				if (((merge_type==7) && (r->type=='s')) ||
						((merge_type==8) && (r->type=='r')))
				{
					in2->refs.tag(r);
					ln2->item = NULL;
				}
			}
			in->refs.tidy();
		}
	}
}

Interface *Topology::lookupInterface(Device *d, Interface *in)
{
	Interface *in2;
	Link *ln;
	int k, dvlan, ivlan;

	if (in==NULL) return NULL;

	for (ln=d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in2 = (Interface*) ln->item;
		if (strcmp(in2->descr, in->descr)==0) return in2;
	}

	for (ln=d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in2 = (Interface*) ln->item;
		dvlan = ivlan = -1;
		if (sscanf(in2->descr, "VLAN %d", &k)==1) dvlan = k;
		if (sscanf(in2->descr, "VLAN-%d", &k)==1) dvlan = k;
		if (sscanf(in2->descr, "Vlan%d", &k)==1) dvlan = k;
		if (sscanf(in2->descr, "Vl%d", &k)==1) dvlan = k;
		if (sscanf(in->descr, "VLAN %d", &k)==1) ivlan = k;
		if (sscanf(in->descr, "VLAN-%d", &k)==1) ivlan = k;
		if (sscanf(in->descr, "Vlan%d", &k)==1) ivlan = k;
		if (sscanf(in->descr, "Vl%d", &k)==1) ivlan = k;
		if ((ivlan!=-1) && (ivlan==dvlan)) return in2;
	}

	for (ln=d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in2 = (Interface*) ln->item;
		if ((!in->mac.zero()) && (in2->mac.compare(in->mac)==1)) return in2;
	}

	return NULL;
}

void Topology::addNextHop()
{
	Interface *in, *newin;
	SubnetExtra *r;
	IPAddress *ip;
	Link *ln, *ln2;

	if ((merge_type == 7) || (merge_type == 3) || (merge_type == 1))
	{
		for (ln=parser->d->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			// printf("in->index = %d %s\n", in->index, in->descr);
			newin = lookupInterface(tmpdev, in);
			if (newin!=NULL)
			{
				// printf("lookupInterface found %d %s\n", newin->index, newin->descr);
				for (ln2=parser->subnets.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					r = (SubnetExtra*) ln2->item;
					if ((r->index == in->index) && (!(r->next.loopback())) &&
							(!r->next.zero()) && (!r->ip.zero()))
					{
						if (r->type == 3)
						{
							add_subnet(r->ip, r->mask);
						}else
						if (r->type == 4)
						{
							// FIXME don't add duplicates as not needed (slow enough already)
							ip = new IPAddress();
							ip->set(r->next);
							newin->nexthops.tag(ip);
						}
					}
				}
			}
		}
	}else
	if (merge_type == 8)
	{
		// we need to move over next hop info from origdev
		for (ln=origdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			newin = lookupInterface(tmpdev, in);
			if (newin!=NULL)
			{
				for (ln2=in->nexthops.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					ip = (IPAddress*) ln2->item;
					newin->nexthops.tag(ip);
					ln2->item = NULL;
				}
				in->nexthops.tidy();
			}
		}
	}
}

void Topology::add_subnet(IPAddress ip, int mask)
{
	Link *ln;
	Subnet *s;
	int got=0;

	if (mask > 30) return;
	for (ln=main->session->data->subnets.get_head(); ln!=NULL; ln=ln->next)
	{
		s = (Subnet*) ln->item;
		if ((s->ip.compare(ip)==0) && (mask == s->mask)) got++; else
		if (overlap(ip, mask, s->ip, s->mask))
		{
			// remove overlapping subnets
			ln->item = NULL;
			delete s;
		}
	}
	main->session->data->subnets.tidy();

	if (!got)
	{
		s = new Subnet();
		s->ip.set(ip);
		s->mask = mask;
		main->session->data->subnets.tag(s);
	}
}

// we look at the next hop values for tmpdev
void Topology::layer3(int hop)
{
	Link *ln, *ln2, *ln3, *ln4, *ln5;
	Interface *in, *in2;
	IPAddress *ip, *ip2;
	Device *d;
	char buf[256];
	int found;

	if ((tmpdev->subtype!=device_router) &&
			(tmpdev->subtype!=device_merged))
		return;

	for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		for (ln2=in->nexthops.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			found = 0;
			ip = (IPAddress*) ln2->item;
			for (ln3=main->session->data->devices.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				d = (Device*) ln3->item;
				if (((d->subtype==device_router)||(d->subtype==device_merged)) && (d != tmpdev))
				{
					for (ln4=d->interfaces.get_head(); ln4!=NULL; ln4=ln4->next)
					{
						in2 = (Interface*) ln4->item;
						for (ln5=in2->ip_addresses.get_head(); ln5!=NULL; ln5=ln5->next)
						{
							ip2 = (IPAddress*) ln5->item;
							if (ip2->compare(*ip)==0)
							{
								if (in->network != in2->network)
								{
									merge_networks(in->network, in2->network);
								}
								found = 1;
							}
						}
					}
				}
			}

			if ((found==0) && (hop>0))
			{
				// FIXME need to make sure that task is not already assigned &&
				// the ip passes the discovery filter

				if (main->gui->discover->match(*ip, hop-1))
				{
					sprintf(buf, "%s %s", ip->getstring(), get_community_list_string(main->client));
					main->scheduler->addTask(buf, 1, 3, hop-1);
				}
			}
		}
	}

	// now reverse check
	for (ln3=main->session->data->devices.get_head(); ln3!=NULL; ln3=ln3->next)
	{
		d = (Device*) ln3->item;
		if (((d->subtype==device_router)||(d->subtype==device_merged)) && (d!=tmpdev))
		{
			for (ln4=d->interfaces.get_head(); ln4!=NULL; ln4=ln4->next)
			{
				in2 = (Interface*) ln4->item;
				for (ln2=in2->nexthops.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					ip = (IPAddress*) ln2->item;
					for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
					{
						in = (Interface*) ln->item;
						for (ln5=in->ip_addresses.get_head(); ln5!=NULL; ln5=ln5->next)
						{
							ip2 = (IPAddress*) ln5->item;
							if (ip2->compare(*ip)==0)
							{
								if (in->network != in2->network)
								{
									merge_networks(in->network, in2->network);
								}
							}
						}
					}
				}
			}
		}
	}

	// now same subnet check
	for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		if (in->network->interfaces.size()<2)
		{
			for (ln3=main->session->data->devices.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				d = (Device*) ln3->item;
				if (((d->subtype==device_router)||(d->subtype==device_merged)) && (d!=tmpdev))
				{
					for (ln4=d->interfaces.get_head(); ln4!=NULL; ln4=ln4->next)
					{
						in2 = (Interface*) ln4->item;
						if (checkInterfacesOnSameSubnet(in, in2)==1)
							merge_networks(in->network, in2->network);
					}
				}
			}
		}
	}
}

int Topology::checkInterfacesOnSameSubnet(Interface *in1, Interface *in2)
{
	Subnet *s1, *s2;
	IPAddress *ip1, *ip2;
	Link *ln, *ln2;

	if (in1 == in2) return 0;

	for (ln=in1->ip_addresses.get_head(); ln!=NULL; ln=ln->next)
	{
		ip1 = (IPAddress*) ln->item;
		s1 = find_subnet(ip1);
		if (s1!=NULL)
		{
			for (ln2=in2->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				ip2 = (IPAddress*) ln2->item;
				s2 = find_subnet(ip2);
				if (s2!=NULL)
				{
					if (s1 == s2)
					{
//						printf("Subnet match %s %s %s", in1->device->name, in1->descr, ip1->getstring());
//						printf("&& %s %s %s\n", in2->device->name, in2->descr, ip2->getstring());
						return 1;
					}
				}
			}
		}
	}
	return 0;
}

gboolean Topology::checkIPAndIntOnSameSubnet(IPAddress *ip1, Interface *in)
{
	Subnet *s1, *s2;
	IPAddress *ip2;
	Link *ln, *ln2;

	s1 = find_subnet(ip1);
	if (s1==NULL) return FALSE;

	for (ln=in->ip_addresses.get_head(); ln!=NULL; ln=ln->next)
	{
		ip2 = (IPAddress*) ln->item;
		s2 = find_subnet(ip2);
		if ((s2!=NULL) && (s1==s2))
		{
			return TRUE;
		}
	}
	return FALSE;
}

Subnet *Topology::find_subnet(IPAddress *ip)
{
	Link *ln;
	Subnet *s;

	for (ln=main->session->data->subnets.get_head(); ln!=NULL; ln=ln->next)
	{
		s = (Subnet*) ln->item;
		if (ip->in_subnet(s->ip, s->mask))
			return s;
	}

	return NULL;
}

void Topology::merge_networks(Network *n1, Network *n2)
{
	Link *ln;
	Interface *in;

	if (n1==n2) return;

	// n1 is the new network
	// n2 is the old network
	// we move everything from n2 to n1

	// move over interfaces
	for (ln=n2->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		if (in!=NULL)
		{
			in->network = n1;
			n1->interfaces.tag(in);
			if (default_network_type==network_auto) n1->subtype = network_auto;
			ln->item = NULL;
		}
	}
	n2->interfaces.tidy();

	main->session->data->networks.del(n2);
	if (main->gui->monitor->viewer->cur == n2)
		main->gui->monitor->viewer->cur = NULL;
	delete n2;			// DND
}

// work out layer 2 for device dd
void Topology::layer2(Device *dd)
{
	Link *ln, *ln2, *ln3, *ln4;
	Interface *in, *in2;
//	IPAddress *ip;
	Device *d;
//	Subnet *s1, *s2;
	Ref *r;

	if (dd == NULL) return;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			in = (Interface*) ln2->item;
			in->switch_list_flag = '0';
		}
	}

	if ((dd->subtype==device_router)||(dd->subtype==device_merged))
	{
		for (ln=dd->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			in->switch_list_flag = '1';
		}
	}else
	{
		for (ln=dd->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			for (ln2=main->session->data->devices.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				d = (Device*) ln2->item;
				if ((d!=NULL) && ((d->subtype==device_router)||(d->subtype==device_merged)) && (d != dd))
				{
					for (ln3=d->interfaces.get_head(); ln3!=NULL; ln3=ln3->next)
					{
						in2 = (Interface*) ln3->item;
						for (ln4=in2->refs.get_head(); ln4!=NULL; ln4=ln4->next)
						{
							r = (Ref*) ln4->item;
							if (r->mac.compare(in->mac)==1)
							{
printf("ADD SW LST forwards %s %s\n", d->name, in2->descr);
								in2->switch_list_flag = '1';
							}
						}
					}
				}
			}
		}
		// now do the reverse
		for (ln=dd->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			for (ln4=in->refs.get_head(); ln4!=NULL; ln4=ln4->next)
			{
				r = (Ref*) ln4->item;

				for (ln2=main->session->data->devices.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					d = (Device*) ln2->item;
					if ((d!=NULL) && ((d->subtype==device_router)||(d->subtype==device_merged)) && (d != dd))
					{
						for (ln3=d->interfaces.get_head(); ln3!=NULL; ln3=ln3->next)
						{
							in2 = (Interface*) ln3->item;
							if (r->mac.compare(in2->mac)==1)
							{
printf("ADD SW LST backwards %s %s\n", d->name, in2->descr);
								in2->switch_list_flag = '1';
							}
						}
					}
				}
			}
		}
		// now check at IP level
/*
		s2 = find_subnet(&(dd->pingname));
		if (s2!=NULL)
		{
			for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
			{
				d = (Device*) ln->item;
				if ((d!=NULL) && ((d->subtype==device_router)||(d->subtype==device_merged)) && (d != dd))
				{
					for (ln3=d->interfaces.get_head(); ln3!=NULL; ln3=ln3->next)
					{
						in2 = (Interface*) ln3->item;
						for (ln4=in2->ip_addresses.get_head(); ln4!=NULL; ln4=ln4->next)
						{
							ip = (IPAddress*) ln4->item;
							s1 = find_subnet(ip);
							if (s1 == s2)
							{
								in2->switch_list_flag = '1';
							}
						}
					}
				}
			}
		}
*/
	}
}

void Topology::layer2hack()
{
	Link *ln, *ln3, *ln4;
	Interface *in, *in2;
	Device *d;

	if ((tmpdev->subtype==device_router)||(tmpdev->subtype==device_merged))
	{
		for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
		{
			in = (Interface*) ln->item;
			for (ln3=main->session->data->devices.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				d = (Device*) ln3->item;
				if (d->subtype==device_switch)
				{
					for (ln4=d->interfaces.get_head(); ln4!=NULL; ln4=ln4->next)
					{
						in2 = (Interface*) ln4->item;
						if (checkInterfacesOnSameSubnet(in, in2)==1)
							merge_networks(in->network, in2->network);
					}
				}
			}
		}
	}else
	if (tmpdev->subtype==device_switch)
	{
		for (ln3=main->session->data->devices.get_head(); ln3!=NULL; ln3=ln3->next)
		{
			d = (Device*) ln3->item;
			if ((d->subtype==device_router)||(d->subtype==device_merged))
			{
				for (ln4=d->interfaces.get_head(); ln4!=NULL; ln4=ln4->next)
				{
					in2 = (Interface*) ln4->item;
					for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
					{
						in = (Interface*) ln->item;
						if (checkInterfacesOnSameSubnet(in, in2)==1)
							merge_networks(in->network, in2->network);
					}
//					if (checkIPAndIntOnSameSubnet(tmpdev->pingname, in2)==TRUE)
//						merge_networks(???, in2->network);
				}
			}
		}
	}
}

void Topology::layer2go()
{
	Link *ln, *ln2;
	Interface *in;
	Device *d;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if ((d!=NULL)&&((d->subtype==device_router)||(d->subtype==device_merged)))
		{
			for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				in = (Interface*) ln2->item;
				if (in->switch_list_flag == '1')
					connectLayer2(in);
			}
		}
	}
}

void Topology::connectLayer2(Interface *in)
{
	Link *ln, *ln2;
	Device *sw;
	Interface *in2, *sin;
	Network *n;

	if (in==NULL) return;

	printf("Building layer 2 for %s %s\n", in->device->name, in->descr);

	buildSwitchList(in);

	for (ln=switch_list.get_head(); ln!=NULL; ln=ln->next)
	{
		in2 = (Interface*) ln->item;
		printf("  Switch list item %s (%c)\n", in2->device->name, in2->device->subtype);
	}

	// first break all networks in the switch list
	for (ln=switch_list.get_head(); ln!=NULL; ln=ln->next)
	{
		sin = (Interface*) ln->item;
		sw = sin->device;
		sw->tricky = 0;
		if (sw!=NULL)
		{
			for (ln2=sw->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				in2 = (Interface*) ln2->item;
				n = in2->network;
				if (n->interfaces.size()>1)
				{
					severNetwork(n);
					main->session->data->networks.del(n);
					if (main->gui->monitor->viewer->cur == n)
						main->gui->monitor->viewer->cur = NULL;
					delete n;
				}
			}
		}
	}

	recLayer2(in);
}

void Topology::severNetwork(Network *nw)
{
	Link *ln;
	Interface *in;
	Network *n;

	for (ln=nw->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		n = new Network();
		n->interfaces.tag(in);
		if (default_network_type==network_auto) n->subtype = network_auto;
		in->network = n;
		ln->item = NULL;
		main->session->data->networks.tag(n);
	}
	nw->interfaces.tidy();
}

void Topology::recLayer2(Interface *in)
{
	Link *ln, *ln2, *ln3, *ln4, *ln5;
	Device *sw, *sw2, *d;
	Interface *sin, *sin2, *in2, *in3;
	List working_list;
	List joins;
	Join *j;
	Ref *r;
	int tag, match;

	if (in->mac.zero()) return;
	if (in->network->interfaces.size()>1) return;

	// start by building a subset of the switch list

//	printf("SUB   Interface %s %s\n", in->device->name, in->descr);
//	printf("SUB   Build subset of switch list\n");

	working_list.empty();
	for (ln=switch_list.get_head(); ln!=NULL; ln=ln->next)
	{
		sin = (Interface*) ln->item;
		d = sin->device;
		if (d->tricky==0)
		{
			match = 0;
			for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				in2 = (Interface*) ln2->item;
				// for each switch interface
				// printf("SUB    Checking %s %s\n", d->name, in2->descr);
				for (ln3 = in2->refs.get_head(); ln3!=NULL; ln3=ln3->next)
				{
					r = (Ref*) ln3->item;
					if (r->mac.compare(in->mac)==1)
					{
						if (match == 0)
						{
							printf("SUB     Working List Switch %s\n", d->name);
							working_list.tag(in2);
							match = 1;
						}
					}
				}
			}
		}
	}

	// now build a joins table for the working list

	printf("SUB     Build Joins List\n");
	joins.empty();
	for (ln=working_list.get_head(); ln!=NULL; ln=ln->next)
	{
		sin = (Interface*) ln->item;
		sw = sin->device;
//		printf("SUB     JOIN   Device %s\n", sw->name);

		if (sw!=NULL)
		{
			for (ln2=sw->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				in2 = (Interface*) ln2->item;
//				printf("SUB     JOIN   Interface %s %s %s\n", sw->name, in2->descr, in2->mac.getstring());
				if ((!in2->mac.zero()) && (in2->type!=53))	// also sc0 & VL1
				{
					// we have a MAC Address to check for



					for (ln3=working_list.get_head(); ln3!=NULL; ln3=ln3->next)
					{
						sin2 = (Interface*) ln3->item;
						sw2 = sin2->device;
						if ((sw2!=NULL) && (sw!=sw2))
						{
							for (ln4=sw2->interfaces.get_head(); ln4!=NULL; ln4=ln4->next)
							{
								in3 = (Interface*) ln4->item;
								for (ln5=in3->refs.get_head(); ln5!=NULL; ln5=ln5->next)
								{
									r = (Ref*) ln5->item;
									if (r->mac.compare(in2->mac)==1)
									{
//										printf("SUB JOIN Found %s %s ==== %s %s\n",
//											in2->device->name, in2->descr, 
//											in3->device->name, in3->descr);
										j = new Join();
										j->in1 = in2;
										j->in2 = in3;
										j->tag = 0;
										joins.tag(j);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	// print joins list
	for (ln=joins.get_head(); ln!=NULL; ln=ln->next)
	{
		j = (Join*) ln->item;
		printf("JOIN %s %s <-> %s %s\n",
			j->in1->device->name, j->in1->descr, 
			j->in2->device->name, j->in2->descr);
	}

	// Now go through the list of joins and look for interfaces that match those
	// in the switch list. Since the entries in the switch list all point to the
	// router interface there will be one which does not match which is the root
	// of the tree. If there is more than one then there is a hub or other
	// device in the way and we can't build an accurate tree. We then recursively
	// repeat for each matching device until complete.

	for (ln=working_list.get_head(); ln!=NULL; ln=ln->next)
	{
		in2 = (Interface*) ln->item;
		if ((in2!=NULL) && (in2->device!=NULL))
		{
			tag = 0;
			for (ln2=joins.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				j = (Join*) ln2->item;
				if ((j->tag == 0) && ((in2 == j->in1) || (in2 == j->in2)))
				{
					tag = 1;
				}
			}
			if (tag == 0)
			{
				merge_networks(in->network, in2->network);
				in2->device->tricky = 1;
				printf("SUB   Merging in %s\n", in2->device->name);
			}
		}
	}

	for (ln=working_list.get_head(); ln!=NULL; ln=ln->next)
	{
		in2 = (Interface*) ln->item;
		if ((in2!=NULL) && (in2->device!=NULL) && (in2->device->tricky == 1))
		{
			for (ln2=in2->device->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				in3 = (Interface*) ln2->item;
				if ((in2!=in3) && (in3->network->interfaces.size()<2) &&
						(in3->type!=53))	// also sc0 & VL1
				{
					printf("RECURSIVE CHECK %s %s\n", in3->device->name, in3->descr);
					recLayer2(in3);
				}
			}
			in2->device->tricky = 2;
		}
	}

	// wipe the joins list
	for (ln=joins.get_head(); ln!=NULL; ln=ln->next)
	{
		j = (Join*) ln->item;
		if (j!=NULL) delete j;
	}
	joins.empty();
}

void Topology::buildSwitchList(Interface *in2)
{
	Link *ln, *ln2, *ln3;
	Interface *in;
	Device *d;
	Ref *r;
	int match;

	if (!in2->mac.zero())
	{
		// clear the existing switch list
		switch_list.empty();

		// build the new switch list
		for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
		{
			d = (Device*) ln->item;
			if (d->subtype==device_switch)
			{
				match = 0;
				for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					in = (Interface*) ln2->item;

					for (ln3 = in2->refs.get_head(); ln3!=NULL; ln3=ln3->next)
					{
						r = (Ref*) ln3->item;
						if (r->mac.compare(in->mac)==1)
						{
							if (match == 0)
							{
								switch_list.tag(in);
								match = 1;
							}
						}
					}
/*
					if ((checkInterfacesOnSameSubnet(in, in2)==1) && (match == 0))
					{
						switch_list.tag(in);
						match = 1;
					}
*/
					// and try it the other way
					for (ln3 = in->refs.get_head(); ln3!=NULL; ln3=ln3->next)
					{
						r = (Ref*) ln3->item;
						if (r->mac.compare(in2->mac)==1)
						{
							if (match == 0)
							{
								switch_list.tag(in);
								match = 1;
							}
						}
					}
				}
			}
		}
	}
}

Interface *Topology::isMACOnInterface(MACAddress mac)
{
	Link *ln, *ln2;
	Device *d;
	Interface *in;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			in = (Interface*) ln2->item;
			if (mac.compare(in->mac)==1)
				return in;
		}
	}

	return NULL;
}

Machine *Topology::isMACAMachine(MACAddress mac)
{
	Link *ln;
	Machine *m;

	for (ln=main->session->data->machines.get_head(); ln!=NULL; ln=ln->next)
	{
		m = (Machine*) ln->item;
		if (mac.compare(m->mac)==1)
			return m;
	}

	return NULL;
}

void Topology::copyOverMachines()
{
	Link *ln, *ln2;
	Interface *in, *in2, *in3;
	Machine *m, *m2, *m3;

	for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		if ((in->type != 24) && (in->type != 53))			// virtual OR loopback
		{
			in2 = lookupInterface(parser->d, in);
			if (in2!=NULL)
			{
				for (ln2=in2->machines.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					m = (Machine*) ln2->item;

					in3 = isMACOnInterface(m->mac);
					if (in3==NULL)
					{
						m2 = isMACAMachine(m->mac);
						if (m2==NULL)
						{
							// add machine here
							m3 = new Machine();
							m3->ip.set(m->ip);
							m3->mac.set(m->mac);
							m3->last_seen = parser->d->last_seen;
							m3->interface = in;
							in->machines.tag(m3);
							main->session->data->machines.tag(m3);
						}else
						{
							// m2 existing machine (m2->interface)
							// m3 new machine (in)
							if (newLocationBest(m2->interface, in)==TRUE)
							{
								m2->interface->machines.del(m2);
								m2->interface = in;
								m2->interface->machines.tag(m2);
							}
							m2->last_seen = parser->d->last_seen;
							if (m3->ip.zero()) m3->ip.set(m2->ip);
						}
					}
				}
			}
		}
	}
}

gboolean Topology::newLocationBest(Interface *oldin, Interface *newin)
{
	if (oldin->network->interfaces.size() > newin->network->interfaces.size())
		return TRUE;
	return FALSE;
}




/*
	Link *ln, *ln2, *ln3;
	Interface *oin, *nin;
	Machine *m, *m2;
	int result;

	for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		nin = (Interface*) ln->item;
		if ((nin->type != 24) &&		// loopback
				(nin->type != 53))			// virtual
		{
			oin = lookupInterface(parser->d, nin);
			if (oin!=NULL)
			{
				for (ln2=oin->machines.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					m = (Machine*) ln2->item;
					if (m->tag==0)
					{
						nin->machines.tag(m);
						main->session->data->machines.tag(m);
						m->interface = nin;

						// now we need to look for duplicate machines

						for (ln3=main->session->data->machines.get_head(); ln3!=NULL; ln3=ln3->next)
						{
							m2 = (Machine*) ln3->item;
							if (m2->mac.compare(m->mac)==1)
							{
								printf("DUPLICATE MACHINE on %s %s AND %s %s\n",
									m->interface->device->name, m->interface->descr,
									m2->interface->device->name, m2->interface->descr);
								result = pickMachine(m, m2);
								if (result==1)
								{
									// use new, copy over user defined values
									strcpy(m->name, m2->name);
									if (m->ip.zero()) m->ip.set(m2->ip);
									strcpy(m->location, m2->location);
									strcpy(m->descr, m2->descr);
									strcpy(m->contact, m2->contact);
//									main->session->data->machines.del(m2);
								}else
								{
									// keep orig, copy over timestamp
									if (m2->ip.zero()) m2->ip.set(m->ip);
									m2->last_seen.ts = m->last_seen.ts;
//									main->session->data->machines.del(m);
								}
// FIXME if delete entry in for loop above need to compensate
							}
						}
					}
				}
			}
		}
	}
*/

int Topology::pickMachine(Machine *m1, Machine *m2)
{
/*
	int t1, t2;

	t1 = t2 = 0;
	if ((descrVlan(m1->interface)==1) || (m1->interface->type == 53)) t1 = 1;
	if ((descrVlan(m2->interface)==1) || (m2->interface->type == 53)) t2 = 1;
	if ((t1==0) && (t2==1)) return 1;
	if ((t1==1) && (t2==0)) return 2;
*/

	return 1;
}

void Topology::deleteTaggedDevices()
{
	Link *ln;
	Device *d;

	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if ((d!=NULL) && (d->tag==1))
		{
			deleteSingleDevice(d);
			ln->item = NULL;
		}
	}
	main->session->data->devices.tidy();
}

void Topology::deleteSingleDevice(Device *d)
{
	Link *ln2, *ln3;
	Interface *in;
	IPAddress *ip;
	Machine *m;
	Network *n;
	Ref *r;

	if (d==NULL) return;

	switch_list.empty();
	for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
	{
		in = (Interface*) ln2->item;
		if (in != NULL)
		{
			n = in->network;
			n->interfaces.del(in);
			if (default_network_type==network_auto) n->subtype = network_auto;
			if (n->interfaces.size()==0)
			{
				main->session->data->networks.del(n);
				if (main->gui->monitor->viewer->cur == n)
					main->gui->monitor->viewer->cur = NULL;
				delete n;			// DND
			}

			for (ln3=in->refs.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				r = (Ref*) ln3->item;
				if (r!=NULL) delete r;
			}
			in->refs.empty();

			for (ln3=in->nexthops.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				ip = (IPAddress*) ln3->item;
				if (ip!=NULL) delete ip;
			}
			in->nexthops.empty();

			for (ln3=in->ip_addresses.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				ip = (IPAddress*) ln3->item;
				if (ip!=NULL) delete ip;
			}
			in->ip_addresses.empty();

			for (ln3=in->machines.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				m = (Machine*) ln3->item;
				main->session->data->machines.del(m);
				if (main->gui->monitor->viewer->cur == m)
					main->gui->monitor->viewer->cur = NULL;
				delete m;			// DND
			}
			in->machines.empty();
		}
		ln2->item = NULL;
		delete in;
	}
	d->interfaces.tidy();

	if (main->gui->monitor->viewer->cur == d)
		main->gui->monitor->viewer->cur = NULL;
	delete d;			// DND
}

